package require edit 1.0
#--------------------------------
#	menu for linemap
#--------------------------------
menu .lmapmenu -tearoff no
.lmapmenu add command -label "Previous Marked line" -command [list Jump2MarkedLine backwards]
.lmapmenu add command -label "Next Marked line" -command [list Jump2MarkedLine forward]
.lmapmenu add separator
.lmapmenu add command -label "Copy all marked lines" -command "TreatMarkedLines copy"
.lmapmenu add command -label "Cut all marked lines" -command "TreatMarkedLines cut"
.lmapmenu add command -label "Delete all marked lines" -command "TreatMarkedLines delete"
.lmapmenu add separator
.lmapmenu add command -label "Clear All marks" -command "ClearAllMarks"
#--------------------------------
proc TreatMarkedLines {{command copy}} {
	set txt .f.t
	set old [list]
	set clip ""
	set flg 0
	foreach idx [edit::getAllMarkedIndex $txt 1.0 end] {
		set flg 1
		set start [$txt index "$idx linestart"]
		if {[lsearch $old $start] > -1} {continue}
		lappend $old $start
		switch -- $command {
			copy {
				append clip [$txt get $start "$start lineend"]\n
			}
			cut {
				append clip [$txt get $start "$start lineend"]\n
				$txt delete $start "$start lineend"
			}
			delete {
				$txt delete $start "$start lineend"
			}
		}
	}
	if {$flg == 0} {
		bell
		return
	}
	if {$command != "delete"} {
		clipboard clear
		clipboard append $clip
	}
	#delete all marks
	if {$command != "copy"} {
		foreach m [$txt mark names] {
			$txt mark unset $m
		}
	}
	$txt refresh
	return
}
#--------------------------------
proc ClearAllMarks {} {
	set txt .f.t
	set marked [edit::getAllMarkedIndex $txt 1.0 end]
	if {!$marked} {
		bell
		return
	}
	foreach m [$txt mark names] {
		$txt mark unset $m
	}
	$txt refresh
	return
}
#--------------------------------
proc Jump2MarkedLine {direction} {
	set txt .f.t
	set index [edit::getMarkedLine $txt insert $direction]
	if {$index == ""} {
		bell
		return
	}
	$txt mark set insert $index
	$txt see insert
	$txt tag remove sel 1.0 end
	$txt tag add sel "$index linestart" "$index lineend"
}
#--------------------------------
proc main {} {
pack [frame .f] -fill both -expand 1
pack [scrollbar .f.s -command {.f.t yview}] -side right -fill y
set tabw "[expr {4 * [font measure {Courier 12} 0]}] left"
set txt [edit .f.t -font {Courier 12} -yscrollcommand {.f.s set} \
	-highlightbrackets 1 -relief flat -padx 4 -tabs $tabw -undo 1]
pack $txt -fill both -expand 1
#Generate highlight rules from text file
edit::generateHighlightRule tcl.txt tcl.dat
$txt configure -highlight 1 -highlightrule tcl.dat
#define text tag
set tagfile [file join [file dirname [info script]] tag.txt]
set tags [edit::defineTextTag $tagfile]
edit::applyTextTags $txt $tags
#add regexp class for highlight
set url_rgx {https?:\/\/[-_.!~*'()a-zA-Z0-9;?:\@&=+\$,%#]+}
edit::addHighlightRgx $txt $url_rgx url; #url-tag is defined in $tagfile.
set test "http://www.ctrans.org/"
#popup-menu
$txt configure -linemapmenu .lmapmenu

bind all <Control-F10> {console show;break}
$txt insert end [info body main]
}
main