#--------------------------------
# edit.tcl
#--------------------------------
package require hashtrie 1.0
package provide edit 1.0
package require unicodeblock 1.1

namespace eval edit {
	variable lmapSelectFirst ""
}

#--------------------------------
#	win is used as a unique token to create arrays for each instance
#--------------------------------
proc edit::getAr {win suffix name} {
	set arName __edit[set win][set suffix]
	uplevel [list upvar #0 $arName $name]
	return $arName
}
#--------------------------------
proc edit {w args} {
	if {[llength $args] & 1} {
		return -code error "invalid number of arguments given to editor (uneven number after window) : $args"
	}
	edit::getAr $w config settings
	set tmp [text .__temptxt]
	set settings(-foreground) [$tmp cget -foreground]
	set settings(-background) [$tmp cget -background]
	set settings(-font) [$tmp cget -font]
	set settings(-relief) [$tmp cget -relief]
	destroy $tmp
	# default setting
	set settings(-padx) 0
	set settings(-pady) 0
	set settings(-rmargin) 0
	set settings(-yscrollcommand) ""
	set settings(-linemap) 1
	set settings(-linemapfg) #BBBBBB
	set settings(-linemapbg) #F2F2F2
	set settings(-linemapwidth) 35
	#set settings(-linemapbd) 1
	set settings(-linemap_mark_command) ""
	set settings(-linemapmenu) ""
	set settings(-linemap_markable) 1
	set settings(-linemap_select_fg) #ff0000
	set settings(-linemap_separator) 0
	set settings(-linemap_separator_bg) #ffffff
	set settings(-highlight) 0
	set settings(-highlightrule) "" ; #See also edit::setHighlightRule
	set settings(-highlightbrackets) 0
	set settings(-nocase) 0
	set settings(-brackets) "\[\]<>{}()"
	set settings(-autoindent) 0
	set settings(-spacing1) 5
	set settings(-spacing2) 5
	set settings(-spacing3) 5
	set settings(-relief) flat
	set settings(win) $w
	set settings(modified) 0

	# overwrite
	array set settings $args

	# special option
	set settings(flags) [list -yscrollcommand -linemap -linemapfg -linemapbg \
	-font -linemap_mark_command -highlight -linemap_markable -linemap_select_fg \
	-highlightbrackets -brackets -autoindent -nocase \
	-linemap_select_font -linemapwidth -background -highlightrule \
	-linemapmenu -linemap_separator -linemap_separator_bg -rmargin]

	#Now remove flags that will confuse text and those that need modification:
	foreach arg $settings(flags) {
		if {[set loc [lsearch $args $arg]] >= 0} {
			set args [lreplace $args $loc [expr {$loc + 1}]]
		}
	}

	# base frame
	frame $w -background $settings(-background) -pady $settings(-pady) -highlightthickness 0 -class edit;
	pack $w -fill both -expand 1
	set settings(-linemap_select_font) [linsert $settings(-font) end bold underline]
	set settings(-pady) 0

	# create linemap
	canvas $w.lmap -width $settings(-linemapwidth) -background $settings(-linemapbg) \
		 -highlightthickness 0 -takefocus 0
	set topWin [winfo toplevel $w]
	bindtags $w.lmap [list $w.lmap $topWin all]
	
	if {$settings(-linemap) == 1} {
		pack $w.lmap -side left -fill y
	}

	#scroll
	set args [concat $args [list -yscrollcommand \
		[list edit::event:yscroll $w $settings(-yscrollcommand)]]]

	# create separator
	frame $w.separator -width 1 -background $settings(-linemap_separator_bg) \
		-bd 1 -highlightthickness 0 -takefocus 0
	if {$settings(-linemap) == 1 && $settings(-linemap_separator) == 1} {
		pack $w.separator -side left -fill y
	}

	# create main text widget
	if {$settings(-rmargin) == 0} {
		pack [eval text \$w.txt -font \$settings(-font) -background $settings(-background) $args] -side left -fill both -expand yes
	} else {
		pack [eval text \$w.txt -font \$settings(-font) -background $settings(-background) $args] -side left -fill y
	}

	# binds(txt)
	bind $w.txt <Configure> [list edit::linemapUpdate $w]
	bind $w.txt <Control-MouseWheel> {edit::textFontZoom %W %D}
	bind $w.txt <KeyRelease> [list edit::highlightBrackets $w]
	bind $w.txt <ButtonRelease> [list edit::highlightBrackets $w]
	bind $w.txt <Return> {edit::autoIndent %W ; break}
	bind $w.txt <Control-Up> {%W yview scroll -2 units ; break}
	bind $w.txt <Control-Down> {%W yview scroll 2 units ; break}
	bind $w.txt <Double-1> {
		set cur [::tk::TextClosestGap %W %x %y]
		set c [%W get $cur]
		set class [unicodeblock::getblock $c]
		set first [edit::getTextPrevPos %W $cur $class edit::getWordBreakBefore]
		set last [edit::getTextNextPos %W [%W index $cur+1c] $class edit::getWordBreakAfter]
		%W tag remove sel 1.0 end
		%W tag add sel $first $last
		update idletasks
		catch {%W mark set insert sel.last}
		break
	}
	bind $w.txt <Control-Left> {
		tk::TextSetCursor %W [%W index "insert-1c"]
		set class [unicodeblock::getblock [%W get insert]]
		tk::TextSetCursor %W [edit::getTextPrevPos %W insert $class edit::getWordBreakBefore]
		break
	}
	bind $w.txt <Control-Right> {
		set class [unicodeblock::getblock [%W get insert]]
		tk::TextSetCursor %W [%W index "insert+1c"]
		tk::TextSetCursor %W [edit::getTextNextPos %W insert $class edit::getWordBreakAfter]
		break
	}
	bind $w.txt <Shift-Control-Left> {
		set class [unicodeblock::getblock [%W get insert-1c]]
		tk::TextKeySelect %W [edit::getTextPrevPos %W [%W index insert-1c] $class edit::getWordBreakBefore]
		break
	}
	bind $w.txt <Shift-Control-Right> {
		set class [unicodeblock::getblock [%W get insert]]
		tk::TextKeySelect %W [edit::getTextNextPos %W insert $class edit::getWordBreakAfter]
		break
	}
	# binds(lmap)
	bind $w.lmap <ButtonPress-1> [list edit::lmapSelectStart $w %y]
	bind $w.lmap <ButtonPress-3> [list edit::lmapSelectStart $w %y]
	bind $w.lmap <B1-Motion> [list edit::lmapSelectMove $w %y]
	bind $w.lmap <ButtonRelease-1> [list edit::lmapSelectEnd $w %y]
	bind $w.lmap <Double-1> [list edit::linemapToggleMark $w %y]
	bind $w.lmap <Shift-1> [list edit::lmapSelectTo $w %y]
	# binds(frame)
	bind $w <ButtonPress-1> [list edit::clickMargin $w %y]
	bind $w <FocusIn> [list focus -force $w.txt]

	if {$settings(-linemapmenu) != ""} {
		bind $w.lmap <ButtonRelease-3> [list edit::linemapPopup %X %Y $settings(-linemapmenu)]
	}

	rename $w _dummy_$w
	rename $w.txt $w._txt

	bind $w <Destroy> [list edit::event:Destroy $w %W]
	bindtags $w.txt [linsert [bindtags $w.txt] 0 $w]

	interp alias {} $w {} edit::instanceCmd $w
	interp alias {} $w.txt {} $w

	edit::modified $w 0
	edit::buildArgParseTable $w
	if {$settings(-highlight) == 1 && $settings(-highlightrule) !=""} {
		edit::setHighlightRule $w $settings(-highlightrule)
	}
	if {$settings(-highlightbrackets) == 1 && $settings(-brackets) !=""} {
		edit::defineBrackets $w $settings(-brackets)
	}
	edit::initHighlightRgx $w
	return $w
}
#--------------------------------
proc edit::getTextNextPos {w start cur_class op} {
	set str [$w get $start "$start lineend"]
	set pos [$op $str $cur_class]
	if {$pos > -1} {
		return [$w index "$start + $pos chars"]
	}
	return [$w index "$start lineend"]
}
#--------------------------------
proc edit::getWordBreakAfter {str cur_class} {
	set pos 0
	foreach c [split $str ""] {
		if {[unicodeblock::getblock $c] != $cur_class} {
			return $pos
		}
		incr pos
	}
	return -1
}
#--------------------------------
proc edit::getTextPrevPos {w end cur_class op} {
	set str [$w get "$end linestart" $end]
	set pos [$op $str $cur_class]
	if {$pos > -1} {
		return [$w index "$end - $pos chars"]
	}
	return [$w index "$end linestart"]
}
#--------------------------------
proc edit::getWordBreakBefore {str cur_class} {
	set pos 0
	for {set i [expr [string length $str]-1]} {$i > -1} {incr i -1} {
		set c [string index $str $i]
		if {[unicodeblock::getblock $c] != $cur_class} {
			return $pos
		}
		incr pos
	}
	return -1
}
#--------------------------------
#	Textzoom
#	thanks: http://wiki.tcl.tk/15868
#--------------------------------
proc edit::textFontZoom {w D} {
	set font [$w cget -font]
	set fname [lindex $font 0]
	set fsize [lindex $font 1]
	if { $D >= 0 } {
		incr fsize 2
	} else {
		incr fsize -2
	}
	set font [list $fname $fsize]
	$w configure -font $font
	$w see insert
}
#--------------------------------
#	popup menu
#--------------------------------
proc edit::linemapPopup {x y menu} {
	tk_popup $menu $x $y
	return
}
#--------------------------------
#	line select
#--------------------------------
proc edit::lmapSelectStart {w y} {
	variable lmapSelectFirst
	focus .
	focus $w
	$w._txt tag remove sel 1.0 end
	set idx [$w._txt index @0,$y]
	set lmapSelectFirst [$w._txt index "$idx linestart"]
	$w._txt tag add sel "$idx linestart" "$idx lineend"
	$w._txt mark set insert $lmapSelectFirst
}
#--------------------------------
proc edit::lmapSelectTo {w y} {
	variable lmapSelectFirst
	set idx [$w._txt index @0,$y]
	$w._txt tag remove sel 1.0 end
	if {[$w._txt compare $idx > insert]} {
		set start insert
		set end [$w._txt index "$idx lineend"]
	} else {
		set start [$w._txt index "$idx linestart"]
		set end "insert"
	}
	set lmapSelectFirst insert
	$w._txt tag add sel $start $end
}
#--------------------------------
proc edit::lmapSelectMove {w y} {
	variable lmapSelectFirst
	if {$lmapSelectFirst == ""} {return}
	set idx [$w._txt index @0,$y]
	$w._txt tag remove sel 1.0 end
	if {$y < 0} {
		$w yview scroll -1 units
	} elseif {[winfo height $w] < $y} {
		$w yview scroll 1 units
	}
	if {[$w._txt compare $lmapSelectFirst > [$w._txt index "$idx linestart"]]} {
		set start [$w._txt index "$idx linestart"]
		set end [$w._txt index "$lmapSelectFirst lineend"]
	} else {
		set start $lmapSelectFirst
		set end [$w._txt index  "$idx lineend"]
	}
	$w._txt tag add sel $start $end
}
#--------------------------------
proc edit::lmapSelectEnd {w y} {
	variable lmapSelectFirst
	if {$lmapSelectFirst == ""} {return}
	set idx [$w._txt index @0,$y]
	$w._txt tag remove sel 1.0 end
	if {[$w._txt compare $lmapSelectFirst > [$w._txt index "$idx linestart"]]} {
		set start [$w._txt index "$idx linestart"]
		set end [$w._txt index "$lmapSelectFirst lineend"]
	} else {
		set start $lmapSelectFirst
		set end [$w._txt index  "$idx lineend"]
	}
	$w._txt tag add sel $start $end
	set lmapSelectFirst ""
}
#--------------------------------
#	destroy
#--------------------------------
proc edit::event:Destroy {win dWin} {
	if {![string equal $win $dWin]} {
		return
	}
	catch {rename $win {}}
	interp alias {} $win.txt {}
	array unset [edit::getAr $win config ar]
}
#--------------------------------
#This stores the arg table within the config array for each instance.
#It's used by the configure instance command.
#--------------------------------
proc edit::buildArgParseTable win {
	set argTable [list]

	lappend argTable {1 true yes} -autoindent {
		set configAr(-autoindent) 1
		break
	}

	lappend argTable {0 false no} -autoindent {
		set configAr(-autoindent) 0
		break
	}

	lappend argTable {1 true yes} -nocase {
		set configAr(-nocase) 1
		break
	}

	lappend argTable {0 false no} -nocase {
		set configAr(-nocase) 0
		break
	}

	lappend argTable any -linemap_mark_command {
		set configAr(-linemap_mark_command) $value
		break
	}

	lappend argTable {1 true yes} -linemap {
		if {$configAr(-linemap_separator) == 1} {
			catch {pack $self.separator -before $self.txt -side left -fill y}
		}
		if {[winfo exists $self.separator]} {
			pack $self.lmap -before $self.separator -side left -fill y
		} else {
			pack $self.lmap -before $self.txt -side left -fill y
		}
		set configAr(-linemap) 1
		break
	}

	lappend argTable {0 false no} -linemap {
		pack forget $self.lmap
		set configAr(-linemap) 0
		if {[winfo exists $self.separator]} {
			pack forget $self.separator
		}
		break
	}

	lappend argTable any -linemapmenu {
		if {$value == ""} {
			bind $self.lmap <ButtonPress-3> {break}
		} else {
			bind $self.lmap <ButtonPress-3> [list edit::linemapPopup %X %Y $value]
		}
		set configAr(-linemapmenu) $value
		break
	}

	lappend argTable any -yscrollcommand {
		set cmd [list $self._txt config -yscrollcommand [list edit::event:yscroll $self $value]]
		if {[catch $cmd res]} {
			return $res
		}
		set configAr(-yscrollcommand) $value
		break
	}

	lappend argTable any -linemapfg {
		if {[catch {winfo rgb $self $value} res]} {
			return -code error $res
		}
		set configAr(-linemapfg) $value
		break
	}

	lappend argTable any -linemapbg {
		if {[catch {winfo rgb $self $value} res]} {
			return -code error $res
		}
		$self.lmap config -background $value
		set configAr(-linemapbg) $value
		break
	}

	lappend argTable any -linemap_select_fg {
		if {[catch {winfo rgb $self $value} res]} {
			return -code error $res
		}
		set configAr(-linemap_select_fg) $value
		break
	}

	lappend argTable {0 false no} -rmargin {
		pack configure $self.txt -fill both -expand 1
		set configAr(-rmargin) 0
		break
	}
	lappend argTable {1 true yes} -rmargin {
		pack configure $self.txt -fill y -expand 0
		set configAr(-rmargin) 1
		break
	}

	lappend argTable {1 true yes} -linemap_separator {
		if {![winfo exists $self.linemap_separator]} {
			pack $self.separator -before $self._txt -fill y -side left
		}
		set configAr(-linemap_separator) 1
		break
	}

	lappend argTable {0 false no} -linemap_separator {
		if {[winfo exists $self.linemap_separator]} {
			pack forget $self.separator
		}
		set configAr(-linemap_separator) 0
		break
	}

	lappend argTable any -linemap_separator_bg {
		if {[catch {winfo rgb $self $value} res]} {
			return -code error $res
		}
		$self.separator configure -background $value
		set configAr(-linemap_separator_bg) $value
		break
	}

	lappend argTable any -font {
		$self._txt config -font $value
		set configAr(-font) $value
		set configAr(-linemap_select_font) [concat $value bold underline]
		break
	}
	lappend argTable any -background {
		if {[catch {winfo rgb $self $value} res]} {
			return -code error $res
		}
		_dummy_$self config -background $value
		$self._txt config -background $value
		set configAr(-background) $value
		break
	}

	lappend argTable any -pady {
		_dummy_$self config -pady $value
		set configAr(-pady) $value
		break
	}

	lappend argTable any -highlightrule {
		set configAr(-highlightrule) $value
		edit::setHighlightRule $self $value
		break
	}

	lappend argTable {0 false no} -highlight {
		set configAr(-highlight) 0
		break
	}

	lappend argTable {1 true yes} -highlight {
#		if {$configAr(-highlightrule) == ""} {
#			return -code error "-highlightrule is blank"
#		}
		set configAr(-highlight) 1
		break
	}

	lappend argTable any -brackets {
		set configAr(-brackets) $value
		edit::defineBrackets $self $value
		break
	}

	lappend argTable {1 true yes} -highlightbrackets {
		set configAr(-highlightbrackets) 1
		break
	}

	lappend argTable {0 false no} -highlightbrackets {
		set configAr(-highlightbrackets) 0
		break
	}

	lappend argTable {0 false no} -linemap_markable {
		set configAr(-linemap_markable) 0
		break
	}

	lappend argTable {1 true yes} -linemap_markable {
		set configAr(-linemap_markable) 1
		break
	}

	edit::getAr $win config ar
	set ar(argTable) $argTable
}
#--------------------------------
#	instanceCmd
#--------------------------------
proc edit::instanceCmd {self cmd args} {
	switch -glob -- $cmd {
		append {
			if {[catch {$self._txt get sel.first sel.last} data] == 0} {
				clipboard append -displayof $self $data
			}
		}

		cget {
			set arg [lindex $args 0]
			edit::getAr $self config configAr
			foreach flag $configAr(flags) {
				if {[string match ${arg}* $flag]} {
					return [set configAr($flag)]
				}
			}
			return [$self._txt cget $arg]
		}

		conf* {
			edit::getAr $self config configAr
			if {0 == [llength $args]} {
				set res [$self._txt configure]
				set del [lsearch -glob $res -yscrollcommand*]
				set res [lreplace $res $del $del]
				foreach flag $configAr(flags) {
					lappend res [list $flag [set configAr($flag)]]
				}
				return $res
			}
			array set flags {}
			foreach flag $configAr(flags) {
				set loc [lsearch $args $flag]
				if {$loc < 0} {
					continue
				}
				if {[llength $args] <= ($loc + 1)} {
					return [set configAr($flag)]
				}
				set flagArg [lindex $args [expr {$loc + 1}]]
				set args [lreplace $args $loc [expr {$loc + 1}]]
				set flags($flag) $flagArg
			}

			foreach {valueList flag cmd} $configAr(argTable) {
				if {[info exists flags($flag)]} {
					foreach valueToCheckFor $valueList {
						set value [set flags($flag)]
						if {[string equal "any" $valueToCheckFor]} $cmd \
						elseif {[string equal $valueToCheckFor [set flags($flag)]]} $cmd
					}
				}
			}
			if {[llength $args]} {
				#we take care of configure without args at the top of this branch
				uplevel 1 [linsert $args 0 $self._txt configure]
			}
		}

		copy {
			tk_textCopy $self
		} 

		cut {
			if {[catch {$self.txt get sel.first sel.last} data] == 0} {
				clipboard clear -displayof $self.txt
				clipboard append -displayof $self.txt $data
				$self delete [$self.txt index sel.first] [$self.txt index sel.last]
				edit::modified $self 1
				edit::linemapUpdate $self
			}
		}

		delete {
			set argsLength [llength $args]
			if {$argsLength == 1} {
				#deal with delete n.n
				set deletePos [$self._txt index [lindex $args 0]]
				$self._txt delete $deletePos
				set lineStart [$self._txt index "$deletePos linestart"]
				set lineEnd [$self._txt index "$deletePos + 1 chars lineend"]
				edit::highlight $self $lineStart $lineEnd
				edit::linemapUpdate $self
			} elseif {$argsLength == 2} {
				#now deal with delete n.n n.n
				set deleteStartPos [lindex $args 0]
				set deleteEndPos [lindex $args 1]
				set data [$self._txt get $deleteStartPos $deleteEndPos]
				set lineStart [$self._txt index "$deleteStartPos linestart"]
				set lineEnd [$self._txt index "$deleteEndPos + 1 chars lineend"]
				eval \$self._txt delete $args
				edit::highlight $self $lineStart $lineEnd
				if {[string first "\n" $data] >= 0} {
					edit::linemapUpdate $self
				}
			} else {
				return -code error "invalid argument(s) sent to $self delete: $args"
			}
			edit::modified $self 1
		}

		fastdelete {
			eval \$self._txt delete $args
			edit::modified $self 1
			edit::linemapUpdate $self
		}
		
		fastinsert {
			eval \$self._txt insert $args
			edit::modified $self 1
			edit::linemapUpdate $self
		}

		fastreplace {
			lassign $args start end text
			$self._txt replace $start $end $text
			edit::linemapUpdate $self
			edit::modified $self 1
		}

		highlight {
			edit::highlight $self [lindex $args 0] [lindex $args 1]
		}

		insert {
			if {[llength $args] < 2} {
				return -code error "please use at least 2 arguments to $self insert"
			}
			set insertPos [lindex $args 0]
			set lineStart [$self._txt index "$insertPos linestart"]
			eval \$self._txt insert $args
			set lineEnd [$self._txt index "insert lineend"]
			after idle [list edit::highlight $self $lineStart $lineEnd]
			$self._txt see insert
			edit::modified $self 1
			edit::linemapUpdate $self
		}

		paste {
			tk_textPaste $self 
			edit::modified $self 1
			edit::linemapUpdate $self
		}

		refresh {
			edit::linemapUpdate $self
			edit::highlight $self 1.0 end
		}

		replace {
			lassign $args start end text
			set length [string length $text]
			set ls [$self._txt index "$start linestart"]
			$self._txt replace $start $end $text
			edit::linemapUpdate $self
			edit::modified $self 1
			set le [$self._txt index "$start+${length}chars lineend"]
			after idle [list edit::highlight $self $ls $le]
		}

		edit {
			set subCmd [lindex $args 0]
			set argsLength [llength $args]
			edit::getAr $self config ar
			if {"modified" == $subCmd} {
				if {$argsLength == 1} {
					return $ar(modified)
				} elseif {$argsLength == 2} {
					set value [lindex $args 1]
					set ar(modified) $value
				} else {
					return -code error "invalid arg(s) to $self edit modified: $args"
				}
			} elseif {"separator" == $subCmd} {
				return [$self._txt edit separator]
			} else {
				set start [$self._txt index "insert linestart"]
				set result [uplevel 1 [linsert $args 0 $self._txt $cmd]]
				set end [$self._txt index "insert lineend"]
				edit::linemapUpdate $self
				edit::highlight $self $start $end
				set ar(modified) [$self._txt edit modified]
				return $result
			}
		}

		default { 
			return [uplevel 1 [linsert $args 0 $self._txt $cmd]]
		}
	}
}
#--------------------------------
#	edit::highlight
#	This proc use package hashtrie
#--------------------------------
proc edit::highlight {txt start end} {
	if {[$txt configure -highlight] == 0} {return}
	set trie [edit::getHighlightRule $txt]
	lassign $trie goto output unum
	set rgxs [edit::getHighlightRgx $txt]
	set nocase [$txt configure -nocase]
	if {$end != "end"} {
		set end [$txt index "$end lineend"]
	}
	#delete all tags
	foreach tag [$txt tag names] {
		$txt tag remove $tag $start $end
	}
	set target [$txt get $start $end]
	set LineNumber [lindex [split $start .] 0]
	foreach line [split $target \n] {
		#result = list of list {key start end}
		set result [hashtrie::getall $trie $line]
		edit::textTagging $txt $line $LineNumber $output $result
		#hignlight for regexp patterns
		if {$rgxs != ""} {
			set ranges [edit::searchHighlightRgx $line $rgxs $nocase]
			edit::textTaggingForRgx $txt $LineNumber $ranges
		}
		incr LineNumber
		update
	}
	return
}
#--------------------------------
#	searchHighlightRgx
#--------------------------------
proc edit::searchHighlightRgx {str rgxs nocase} {
	set ranges [list]
	foreach {rgx class} $rgxs {
		set offset 0
		while {1} {
			if {$nocase} {
				set result [regexp -indices -start $offset -nocase -- $rgx $str range]
			} else {
				set result [regexp -indices -start $offset -- $rgx $str range]
			}
			if {!$result} {break}
			lassign $range start end
			incr end
			lappend ranges [list $start $end $class]
			set offset $end
		}
	}
	return $ranges
}
#--------------------------------
#	TextTaggingForRgx
#	ranges = list {range(=start end) class}
#--------------------------------
proc edit::textTaggingForRgx {txt linenumber ranges} {
	foreach range $ranges {
		lassign $range start end class
		$txt tag add $class "${linenumber}.${start}" "${linenumber}.${end}"
	}
	return
}
#--------------------------------
#	TextTagging
#	targets : list of list {key start end}
#	output : a dict that part of trie for highlight(see also package hashtrie).
#--------------------------------
proc edit::textTagging {txt line linenumber output {targets ""}} {
	if {$targets == ""} {return}
	foreach target $targets {
		lassign $target key start end
		lassign [dict get $output $key] type tag pair
		switch $type {
			word {
				if {[string wordstart $line $start] != $start} {continue}
				if {[string wordend $line $start] != $end} {continue}
				$txt tag add $tag "${linenumber}.${start}" "${linenumber}.${end}"
			}
			noblank {
				if {[string wordstart $line $start] != $start} {continue}
				if {[regexp -start $start -indices -- {[\s\"\[\]\{\}]} $line match]} {
					set end [lindex $match 0]
					$txt tag add $tag "${linenumber}.${start}" "${linenumber}.${end}"
				} else {
					$txt tag add $tag "${linenumber}.${start}" "${linenumber}.0 lineend"
				}
			}
			nobreak {
				if {[string wordstart $line $start] == $start} {
					$txt tag add $tag "${linenumber}.${start}" "${linenumber}.0 lineend"
				}
			}
			pair {
				if {[string wordstart $line $start] != $start} {continue}
				set end [string first $pair $line [expr $start+1]]
				if {$end > -1} {
					incr end [string length $pair]
					$txt tag add $tag "${linenumber}.${start}" "${linenumber}.${end}"
				}
			}
			char {
				$txt tag add $tag "${linenumber}.${start}"
			}
			string -
			default {
				$txt tag add $tag "${linenumber}.${start}" "${linenumber}.${end}"
			}
		}
	}
	return
}
#--------------------------------
#	Auto Indent
#	thanks: http://wiki.tcl.tk/15619
#--------------------------------
proc edit::autoIndent {w} {
	if {![$w cget -autoindent]} {
		$w insert [$w index insert] "\n"
		return
	}
	# First, delete the current selection
	if {[$w tag nextrange sel 1.0] != ""} {
		$w delete sel.first sel.last
	}
	# Find the whitespace at the start of the current line
	set startl [$w index "insert linestart"]
	set endl [lindex [split $startl "."] 0].end
	regexp {^([ \t]*)} [$w get $startl $endl] -> ws
	# Create a newline, insert whitespace
	$w insert [$w index insert] "\n$ws"
	# If necessary, scroll the view so cursor is visible
	$w see [$w index insert]
}
#--------------------------------
# set regexp pattern for highlight
#--------------------------------
proc edit::initHighlightRgx {win} {
	set ref [edit::getAr $win highlightRgx ar]
	set ar(rgx) [list]
}
proc edit::getHighlightRgx {win} {
	set ref [edit::getAr $win highlightRgx ar]
	return $ar(rgx)
}
#proc edit::setHighlightRgx {win rgx class} {
#	set ref [edit::getAr $win highlightRgx ar]
#	lappend ar(rgx) $rgx $class
#}
proc edit::addHighlightRgx {win rgx class} {
	if {[catch {regexp -- $rgx "test"} err]} {
		return 0
	} elseif {$rgx == ""} {
		return 0
	}
	set ref [edit::getAr $win highlightRgx ar]
	lappend ar(rgx) $rgx $class
	return 1
}
proc edit::deleteHighlightRgx {win rgx} {
	if {$rgx == ""} {return}
	set ref [edit::getAr $win highlightRgx ar]
	set new [dict remove $ar(rgx) $rgx]
	set ar(rgx) $new
}
proc edit::clearHighlightRgx {win} {
	edit::getAr $win highlightRgx ar
	set ar(rgx) [list]
}
#--------------------------------
proc edit::getHighlightRule {win} {
	set ref [edit::getAr $win highlightRule ar]
	if {![info exists ar(rule)]} {
		set ar(rule) [hashtrie::new]
	}
	return $ar(rule)
}
#--------------------------------
proc edit::setHighlightRule {win filename} {
	if {![file exists $filename]} {
		return -code error "highlight file not found"
	}
	set ref [edit::getAr $win highlightRule ar]
	set f [open $filename r]
	fconfigure $f -encoding utf-8
	set content [read $f]
	if {[string index $content 0] == "\ufeff"} {
		set content [string range $content 1 end]
	}
	close $f
	set ar(rule) $content
}
#--------------------------------
#	addHighlightWord
#	value = [list hilite-type(eg.char) tagname(eg.halfwidth)]
#--------------------------------
proc edit::addHighlightWord {txt word value} {
	set trie [edit::getHighlightRule $txt]
	set trie [hashtrie::add $trie $word $value]
	set ref [edit::getAr $txt highlightRule ar]
	set ar(rule) $trie
}
#--------------------------------
#	deleteHighlightWord
#--------------------------------
proc edit::deleteHighlightWord {txt word} {
	set trie [edit::getHighlightRule $txt]
	hashtrie::remove trie $word
	set ref [edit::getAr $txt highlightRule ar]
	set ar(rule) $trie
}
#--------------------------------
#	generate highlight rule from text file.
#	arguments:
#	src: source file
#	new: file path for save
#	requirement:
#		file encoding must be utf-8.
#		file must be tsv(tab separated value) file.
#		example: word	class
#--------------------------------
proc edit::generateHighlightRule {src new} {
	set rule [dict create]
	#open source file
	set f [open $src r]
	fconfigure $f -encoding utf-8
	set dat [read $f]
	close $f
	if {[string index $dat 0] == "\ufeff"} {
		set dat [string range $dat 1 end]
	}
	#read datas
	foreach line [split $dat \n] {
		set line [split $line \t]
		set key [lindex $line 0]
		set value [lrange $line 1 end]
		if {$key != ""} {dict set rule $key $value}
	}
	#blank trie
	set trie [hashtrie::new]
	#import data
	set trie [hashtrie::import $trie $rule]
	#save
	return [hashtrie::save $trie $new]
}
#--------------------------------
#	highlight brackets
#--------------------------------
proc edit::defineBrackets {win brackets} {
	set ref [edit::getAr $win brackets ar]
	array unset ar
	foreach {left right} [split $brackets ""] {
		if {$left == "" || $right == ""} {continue}
		if {$left == $right} {continue}
		dict set ar(brackets) $left [list "left" $right]
		dict set ar(brackets) $right [list "right" $left]
	}
	return $ar(brackets)
}
#--------------------------------
proc edit::getHighlightBrackets {win} {
	set ref [edit::getAr $win brackets ar]
	return $ar(brackets)
}
#--------------------------------
proc edit::highlightBrackets {txt} {
	if {[$txt configure -highlightbrackets] == 0} {return}
	$txt tag remove brackets 1.0 end
	set brakets [edit::getHighlightBrackets $txt]
	set char [$txt get insert]
	#check exists
	if {![dict exists $brakets $char]} {
		return
	}
	#check escape
	set prevChar [$txt get "insert - 2 chars"]
	if {[string equal $prevChar "\\"]} {
		return
	}
	lassign [dict get $brakets $char] side target
	if {$side == "left"} {
		set pos [$txt index "insert + 1 chars"]
		set texts [$txt get $pos end]
		set increment 1
		set index 0
	} else {
		set texts [$txt get 1.0 insert]
		set increment -1
		set index [expr [string length $texts] -1]
	}
	set count 1
	set i 1
	while {1} {
		set current [string index $texts $index]
		if {$current == ""} {return}
		
		if {[string equal $current $char]} {
			incr count
		} elseif {[string equal $current $target]} {
			incr count -1
		}
		if {$count == 0} {
			set prev [string index $texts [expr $index -1]]
			if {[string equal $prev "\\"]} {
				incr count
			} else {
				break
			}
		} 
		incr index $increment
		incr i
	}
	if {$increment == 1} {
		set pos [$txt index "insert + ${i}c"]
	} else {
		set pos [$txt index "insert - ${i}c"]
	}
	$txt tag add brackets [$txt index insert]
	$txt tag add brackets $pos
	$txt tag raise brackets
	return $pos
}
#--------------------------------
#	defineTextTag
#	CSSƓl̏̃eLXgt@C
#	text tagp̃f[^ǂݍ݁AdictԂ
#--------------------------------
proc edit::defineTextTag {fn} {
	set tags [dict create]
	set properties [list -foreground -background -underline -overstrike -font]
	#open
	if {![file exists $fn]} {return}
	set f [open $fn r]
	fconfigure $f -encoding utf-8
	set content [read $f]
	close $f
	#sets $content in array
	if {[catch {array set class $content} msg]} {
		puts stderr $msg
		return 0
	}
	#regexp pattern for split
	set rgx {([^:]+):(.+)}
	#read each tag class
	foreach {key value} [array get class] {
		set datas [list]
		set value [string map {\t {} \n {} \u0020 {} \u3000 {}} $value]
		foreach declare [split $value ";"] {
			if {![regexp -- $rgx $declare match property body]} {continue}
			if {$property == "" || $body== ""} {continue}
			if {[lsearch $properties $property] == -1} {continue}
			set datas [concat $datas $property $body]
			if {[llength $datas] > 0} {
				dict set tags $key $datas
			}
		}
	}
	return $tags
}
#--------------------------------
#	ApplyTextTags
#	defineTextTagœǂݍ񂾃^Õf[^
#	Ŏw肵EBWFbgitxtjɓKp
#--------------------------------
proc edit::applyTextTags {txt tags} {
	dict for {tag values} $tags {
		eval $txt tag configure $tag $values
	}
}
#--------------------------------
#	update
#This is a proc designed to be overwritten by the user.
#It can be used to update a cursor or animation while
#the text is being highlighted.
#--------------------------------
proc edit::update {} {}
#--------------------------------
#	linemapToggleMark
#--------------------------------
proc edit::linemapToggleMark {win y} {
	edit::getAr $win config configAr
	if {!$configAr(-linemap_markable)} {
		return
	}
	set idx [$win._txt index @0,$y] 
	set line [lindex [split $idx .] 0]
	set marks [$win._txt dump -mark "$idx linestart" "$idx lineend"]
	set lidx [lsearch $marks lmap*]
	if {$lidx > -1} {
		foreach {mark name pos} $marks {
			if {[string range $name 0 3] == "lmap"} {
				$win._txt mark unset $name
			}
		}
		edit::linemapUpdate $win
		set type unmarked
	} else {
		#edit-widget does not accept blank line
		if {[$win._txt get "$idx linestart" "$idx lineend"] ==""} {
			return
		}
		$win._txt mark set "lmap$idx" "$idx linestart"
		edit::linemapUpdate $win
		set type marked
	}

	if {[string length $configAr(-linemap_mark_command)]} {
		uplevel #0 [linsert $configAr(-linemap_mark_command) end $win $type $line]
	}
}
#--------------------------------
#	edit::getMarkedLine
#--------------------------------
proc edit::getAllMarkedIndex {txt start end} {
	set marks [$txt dump -mark $start $end]
	set marked [list]
	set lidx [lsearch $marks lmap*]
	if {$lidx == -1} {return}
	foreach {mark name index} $marks {
		if {[string range $name 0 3] == "lmap"} {
			lappend marked $index
		}
	}
	return $marked
}

#--------------------------------
#	edit::getMarkedLineIndex
#--------------------------------
proc edit::getMarkedLine {txt index direction} {
	switch $direction {
		backwards {
			set lines [edit::getAllMarkedIndex $txt 1.0 $index-1lines]
			if {$lines == ""} {return}
			set line [lindex $lines end]
		}
		forwards -
		default {
			set lines [edit::getAllMarkedIndex $txt $index+1lines end]
			if {$lines == ""} {return}
			set line [lindex $lines 0]
		}
	}
	return $line
}
#--------------------------------
#	edit::clickMargin
#--------------------------------
proc edit::clickMargin {win y} {
	edit::getAr $win config configAr
	set idx [$win._txt index @0,$y]
	$win._txt mark set insert "$idx display lineend -1char"
	return
}
#--------------------------------
#	scroll
#--------------------------------
proc edit::event:yscroll {win clientData args} {
	edit::linemapUpdate $win
	if {$clientData == ""} {
		return
	}
	uplevel #0 $clientData $args
}
#--------------------------------
#	linemapUpdate
#	args is here because -yscrollcommand may call it
#--------------------------------
proc edit::linemapUpdate {win args} {
	if {[winfo exists $win.lmap] != 1} { 
		return
	}
	#change linemap width with last line number
	lassign [split [$win._txt index end] .] eline epos
	edit::getAr $win config ar
	set lmapwidth [expr [font measure $ar(-font) $eline] + 20]
	if {$ar(-linemapwidth) < $lmapwidth} {
		$win.lmap config -width $lmapwidth
		set ar(-linemapwidth) $lmapwidth
	} else {
		set lmapwidth $ar(-linemapwidth)
	}

	#delete all linemap
	foreach item [$win.lmap find all] {
		$win.lmap delete $item
	}

	set pixel 0
	set lastLine {}
	set border [expr {[$win cget -bd] + [$win cget -highlightthickness] + [$win cget -pady]}]
	set winheight [expr {[winfo height $win.lmap] - 2 * $border}]
	while {$pixel < $winheight} {
		set idx [$win._txt index @0,$pixel]
		lassign [split $idx .] line pos
		lassign [$win._txt dlineinfo $idx] dlx dly dlw dlh dlb
		if {[lindex [split $idx .] 1] == "0"} {
			set lwidth [font measure $ar(-font) $line]
			set x [expr ($lmapwidth-$lwidth) + ($lwidth/2) -4]
			set y [expr $dly + (($dlh + ${ar(-spacing1)})/2)]
			#marked line
			set marks [$win._txt dump -mark "$idx linestart" "$idx lineend"]
			if {[lsearch $marks lmap*] > -1} {
				$win.lmap create text $x $y -text "$line" \
				-font $ar(-linemap_select_font) -fill $ar(-linemap_select_fg)
			} else {
				$win.lmap create text $x $y -text "$line" \
				-font $ar(-font) -fill $ar(-linemapfg)
			}
		}
		incr pixel $dlh
	} 
}
#--------------------------------
#	modified
#--------------------------------
proc edit::modified {win value} {
	edit::getAr $win config ar
	set ar(modified) $value
	event generate $win <<Modified>>
	if {$value == 1 && [$win cget -autoseparators] == 0} {$win edit separator}
	return $value
}